                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 6: partial evaluation   */
                     /*-----------------------------------*/
                     /*               main                */
                     /*-----------------------------------*/

#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SlipSlip.h"

#include "SlipMain.h"

#include "SlipCompile.h"
#include "SlipDictionary.h"
#include "SlipEvaluate.h"
#include "SlipNative.h"
#include "SlipPool.h"
#include "SlipPrint.h"
#include "SlipRead.h"
#include "SlipScan.h"
#include "SlipThread.h"

/*------------------------------ private constants ---------------------------*/

static const UNS_type Minimum_number_of_cells = 1024;

static const TXT_type IMM_error_string = "insufficient memory";

/*------------------------------ private types -------------------------------*/

typedef enum { Initiate_loop  = 0,
               Terminate_loop = 1,
               Abort_loop     = 2 } TRA_type;

typedef jmp_buf EXI_type;

/*------------------------------ private variables ---------------------------*/

static CFN_type Continue_print_function;
static ENV_type Rollback_environment;
static EXI_type Trampoline;

static UNS_type Total_number_of_cells;

/*------------------------------ public variables ----------------------------*/

CHR_type Main_Buffer[Main_Buffer_Size];

SYM_type Main_Anonymous;
SYM_type Main_Begin;
SYM_type Main_Define;
CNT_type Main_Empty_Continuation;
ENV_type Main_Empty_Environment;
PAI_type Main_Empty_Pair;
VEC_type Main_Empty_Vector;
FLS_type Main_False;
SYM_type Main_If;
SYM_type Main_Lambda;
NUL_type Main_Null;
NBR_type Main_One;
SYM_type Main_Quote;
SYM_type Main_Set;
TRU_type Main_True;
USP_type Main_Unspecified;
SYM_type Main_While;
NBR_type Main_Zero;

/*------------------------------ private functions ---------------------------*/

static EXP_type continue_print(EXP_type Value,
                               CNT_type Continuation)
  { Print_Print(Value);
    longjmp(Trampoline,
            Terminate_loop); 
    return Main_Unspecified; }

static NIL_type read_eval_print(NIL_type)
  { EXP_type compiled_expression,
             expression,
             value;
    TRA_type status;
    TXT_type input;
    Slip_Print("\n>>>");
    Slip_Read(&input);
    if (setjmp(Trampoline) == Initiate_loop)
      { expression = Read_Parse(input);
        compiled_expression = Compile_Compile(expression);
        Rollback_environment = Dictionary_Environment;
        Thread_Initialize();
        Thread_Push(Continue_print_function,
                    0);
        if ((status = setjmp(Trampoline)) == Initiate_loop)
          { value = Evaluate_Evaluate(compiled_expression); 
            for (;;)
              value = Evaluate_Continue(value); }
        if (status == Abort_loop)
          Dictionary_Environment = Rollback_environment; }}

/*------------------------------ public functions ----------------------------*/

EXP_type Main_Error(TXT_type Error)
  { Slip_Print(Error);
    longjmp(Trampoline,
            Abort_loop);
    return Main_Unspecified; }

EXP_type Main_Error_Procedure(TXT_type Error ,
                              PRC_type Procedure)
  { SYM_type symbol;
    symbol = Procedure->nam;
    return Main_Error_Symbol(Error,
                             symbol); }

EXP_type Main_Error_Procedure_Vararg(TXT_type Error ,
                                     PRZ_type Procedure)
  { SYM_type symbol;
    symbol = Procedure->nam;
    return Main_Error_Symbol(Error,
                             symbol); }

EXP_type Main_Error_Symbol(TXT_type Error ,
                           SYM_type Symbol)
  { TXT_type text;
    text = Symbol->txt;
    return Main_Error_Text(Error,
                           text); }

EXP_type Main_Error_Tag(TXT_type Error,
                        TAG_type Tag )
  { sprintf(Main_Buffer, "%s: %d", Error, Tag);
    return Main_Error(Main_Buffer); }

EXP_type Main_Error_Text(TXT_type Error,
                         TXT_type Text )
  { sprintf(Main_Buffer, "%s: %s", Error, Text);
    return Main_Error(Main_Buffer); }

NIL_type Main_Fatal_Error(TXT_type Error)
  { Slip_Print(Error);
    exit(0); }

TXT_type Main_Load(TXT_type Name)
  { TXT_type input;
    Slip_Load(Name,
              &input);
    return input; }

NIL_type Main_Print(TXT_type Text)
  { Slip_Print(Text); }

TXT_type Main_Read(NIL_type)
  { TXT_type input;
    Slip_Read(&input);
    return input; }

PAI_type Main_Reverse(PAI_type This)
  { PAI_type next,
             pair,
             previous;
    previous = Main_Empty_Pair;
    for (pair = This;
         is_PAI(pair);
         pair = next)
      { next = pair->cdr;
        pair->cdr = previous;
        previous = pair; }
    return previous; }

EXP_type Main_Terminate(NIL_type)
  { exit(0);  
    return Main_Unspecified; }

/*------------------------------ exported functions --------------------------*/

void Slip_REP(char * Memory,
              int    Size)
  { Total_number_of_cells = Size / Memory_Cell_Size;
    if (Total_number_of_cells < Minimum_number_of_cells)
      Main_Fatal_Error(IMM_error_string);
    Memory_Initialize(Memory,
                      Total_number_of_cells);
    Pool_Initialize();
    Main_Empty_Vector       = make_VEC(0);
    Main_False              = make_FLS();
    Main_Null               = make_NUL();
    Main_One                = make_NBR(1);
    Main_True               = make_TRU();
    Main_Unspecified        = make_USP();
    Main_Zero               = make_NBR(0);
    Main_Empty_Continuation = (CNT_type)Main_Null;
    Main_Empty_Environment  = (ENV_type)Main_Null;
    Main_Empty_Pair         = (PAI_type)Main_Null;
    Main_Anonymous          = Pool_Enter(Main_Anonymous_String );
    Main_Begin              = Pool_Enter(Main_Begin_String     );
    Main_Define             = Pool_Enter(Main_Define_String    );
    Main_If                 = Pool_Enter(Main_If_String        );
    Main_Lambda             = Pool_Enter(Main_Lambda_String    );
    Main_Quote              = Pool_Enter(Main_Quote_String     );
    Main_Set                = Pool_Enter(Main_Set_String       );
    Main_While              = Pool_Enter(Main_While_String     );
    Dictionary_Initialize();
    Read_Initialize();
    Compile_Initialize();
    Evaluate_Initialize();
    Native_Initialize();
    Print_Initialize();

    Continue_print_function = make_CFN(continue_print);
    Slip_Print("cpSlip/c version 6");
    for(;;)
      read_eval_print(); }
